# Generated by Django 2.2.12 on 2020-05-23 20:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0003_remove_type_description_not_null'),
    ]

    operations = [
        migrations.CreateModel(
            name='ItemCategory',
            fields=[
                ('category_id', models.IntegerField(primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=250)),
                ('published', models.BooleanField(default=True)),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.RemoveField(
            model_name='type',
            name='group_id',
        ),
        migrations.CreateModel(
            name='ItemGroup',
            fields=[
                ('group_id', models.IntegerField(primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=250)),
                ('published', models.BooleanField(default=True)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='fittings.ItemCategory')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.AddField(
            model_name='type',
            name='group',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='fittings.ItemGroup'),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-09 11:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0014_serverversion'),
    ]

    operations = [
        migrations.AddField(
            model_name='fitting',
            name='created',
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
        migrations.AddField(
            model_name='fitting',
            name='last_updated',
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AlterField(
            model_name='doctrine',
            name='last_updated',
            field=models.DateTimeField(auto_now=True, null=True),
        ),
    ]



# Generated by Django 3.1.2 on 2021-02-21 01:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0013_alter_doctrine_description'),
    ]

    operations = [
        migrations.CreateModel(
            name='ServerVersion',
            fields=[
                ('id', models.BigIntegerField(primary_key=True, serialize=False)),
            ],
            options={
                'verbose_name': 'ServerVersion',
                'default_permissions': (),
            },
        ),
    ]



# Generated by Django 2.2.14 on 2020-08-15 20:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0010_auto_20200718_2227'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='category',
            options={'default_permissions': (), 'verbose_name': 'Category', 'verbose_name_plural': 'Categories'},
        ),
        migrations.AlterField(
            model_name='fitting',
            name='description',
            field=models.TextField(max_length=500),
        ),
    ]



# Generated by Django 2.2.12 on 2020-06-05 07:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0007_auto_20200605_0735'),
    ]

    operations = [
        migrations.AlterField(
            model_name='unicategory',
            name='color',
            field=models.TextField(default='#FFFFFF', max_length=20),
        ),
    ]



# Generated by Django 2.2.14 on 2020-07-18 22:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
        ('fittings', '0009_auto_20200605_2117'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='UniCategory',
            new_name='Category',
        ),
    ]



from django import template
from django.conf import settings


register = template.Library()


@register.simple_tag
def settings_value(name):
    return getattr(settings, name, "")


@register.simple_tag
def get_dt_lang(language):
    """
    Since django language codes and dataTables language codes are not all the same, we need to map one to the other.
    """
    prefix = 'fittings/dt-i18n/'
    special_codes = {
        'de': 'de-DE',
        'en': 'en-GB',
        'es': 'es-ES',
        'zh-hans': 'zh',
        'fr': 'fr-FR',
        'it': 'it-IT'
    }
    if language in special_codes:
        language = special_codes.get(language)

    return f'{prefix}{language}.json'



# Cog used by https://github.com/pvyParts/allianceauth-discordbot

# Cog Stuff
from discord.ext import commands
from discord.embeds import Embed
from discord.colour import Color

# AA Contexts
from aadiscordbot.cogs.utils.decorators import sender_has_perm
from aadiscordbot.app_settings import get_site_url
from django.core.exceptions import ObjectDoesNotExist

from fittings.models import Fitting, Doctrine, Category
from django.db.models import Q


import re

import logging
logger = logging.getLogger('aadiscordbot.cogs.fittings')


class Fittings(commands.Cog):
    """
    A Cog to respond to AA Fittings Links
    Includes more information as the Discord Embed for AA is of limited Use
    """

    def __init__(self, bot):
        self.bot = bot

    @commands.Cog.listener("on_message")
    @sender_has_perm('fittings.access_fittings')
    async def respond_to_fittings(self, message):
        """
        Listens to ALL Messages, act if they contain AA-Fittings Links
        """
        logger.debug(f"Fittings Cog: Message Received {message.content}")
        pattern = re.compile(r".*\/(fittings)\/(doctrine|fit|cat)\/(\d+)\/")

        if re.search(get_site_url(), message.content) is not None:
            fit_tuples = re.findall(pattern, message.content)
        else:
            logger.debug("Fittings Cog: No Match")
            return

        if len(fit_tuples) > 0:
            await message.edit(suppress=True)
            await message.channel.trigger_typing()
            await message.add_reaction(chr(0x231B))
            logger.debug(f"Fittings Cog: Message Match {message.content}")

            for fit_tuple in fit_tuples:
                if fit_tuple[0] == "fittings":
                    if fit_tuple[1] == "fit":
                        logger.debug(f"Fittings Cog: Fit Detected {fit_tuple[2]}")
                        embed = fitting_details(fit_tuple[2])
                    elif fit_tuple[1] == "doctrine":
                        logger.debug(f"Fittings Cog: Doctrine Detected {fit_tuple[2]}")
                        embed = doctrine_details(fit_tuple[2])
                    elif fit_tuple[1] == "cat":
                        logger.debug(f"Fittings Cog: Category Detected {fit_tuple[2]}")
                        embed = category_details(fit_tuple[2])
                    else:
                        await message.reply("I Detected a fitting link, but not a valid page. Please notify an Admin", mention_author=False)
                await message.reply(embed=embed, mention_author=False)
            await message.clear_reaction(chr(0x231B))
        else:
            logger.debug("Fittings Cog: No Match")
            return

    @commands.command(pass_context=True)
    @sender_has_perm('fittings.access_fittings')
    async def fit(self, ctx, fit_name):
        """
        Returns fits and ships matching the string given
        """
        await ctx.channel.trigger_typing()
        await ctx.message.add_reaction(chr(0x231B))

        fits = Fitting.objects.filter(Q(name__icontains=fit_name) | Q(ship_type__type_name__icontains=fit_name))

        if fits.exists() is True:
            for fit in fits:
                embed = fitting_details(fit.id)
                await ctx.reply(embed=embed, mention_author=False)
        else:
            await ctx.reply("No Fits Found", mention_author=False)

        return await ctx.message.clear_reaction(chr(0x231B))

    @commands.command(pass_context=True)
    @sender_has_perm('fittings.access_fittings')
    async def fit_id(self, ctx, fit_id):
        """
        Return information on a fit ID
        """

        embed = fitting_details(fit_id)

        return await ctx.reply(embed=embed, mention_author=False)

    @commands.command(pass_context=True)
    @sender_has_perm('fittings.access_fittings')
    async def fit_ship(self, ctx, ship_name):
        """
        Returns fittings for a Ship Type
        """
        await ctx.channel.trigger_typing()
        await ctx.message.add_reaction(chr(0x231B))

        fits = Fitting.objects.filter(ship_type__type_name__icontains=ship_name)

        if fits.exists() is True:
            for fit in fits:
                embed = fitting_details(fit.id)
                await ctx.reply(embed=embed, mention_author=False)
        else:
            await ctx.reply("No Fits Found", mention_author=False)

        return await ctx.message.clear_reaction(chr(0x231B))

    @commands.command(pass_context=True)
    @sender_has_perm('fittings.access_fittings')
    async def doctrine(self, ctx, doctrine_name):
        """
        Returns doctrines matching the string given
        """
        await ctx.channel.trigger_typing()
        await ctx.message.add_reaction(chr(0x231B))

        try:
            doctrines = Doctrine.objects.filter(name__icontains=doctrine_name)
        except ObjectDoesNotExist:
            return await ctx.reply("No Doctrine Found", mention_author=False)

        for doctrine in doctrines:
            embed = doctrine_details(doctrine.id)
            await ctx.reply(embed=embed, mention_author=False)

        return await ctx.message.clear_reaction(chr(0x231B))

    @commands.command(pass_context=True)
    @sender_has_perm('fittings.access_fittings')
    async def doctrine_id(self, ctx, doctrine_id):
        """
        Return information on a doctrine ID
        """

        embed = doctrine_details(doctrine_id)

        return await ctx.reply(embed=embed, mention_author=False)

    @commands.command(pass_context=True)
    @sender_has_perm('fittings.access_fittings')
    async def category_id(self, ctx, category_id):
        """
        Return information on a category ID
        """

        embed = category_details(category_id)

        return await ctx.reply(embed=embed, mention_author=False)


def fitting_details(fit_id):
    try:
        fit = Fitting.objects.get(id=fit_id)
    except ObjectDoesNotExist:
        embed = Embed(title="Fitting Not Found")
        return

    embed = Embed(title=f"{fit.ship_type.name}: {fit.name}")
    embed.set_thumbnail(
        url=f"https://images.evetech.net/types/{fit.ship_type_type_id}/render?size=128"
    )
    embed.description = fit.description

    doctrines_value = ''
    for doctrine in Doctrine.objects.filter(fittings=fit).values("name", "id"):
        doctrines_value = f"{doctrines_value}[{doctrine['name']}]({get_site_url()}/fittings/doctrine/{doctrine['id']}/)<br>"
    if doctrines_value == '':
        doctrines_value = 'None'

    category_value = ''
    for category in Category.objects.filter(fittings=fit).values("name", "id"):
        category_value = f"{category_value}[{category['name']}]({get_site_url()}/fittings/cat/{category['id']}/)<br>"
    if category_value == '':
        category_value = 'None'

    embed.add_field(
        name="Doctrines:",
        value=doctrines_value
    )
    embed.add_field(
        name="Categories",
        value=category_value
    )
    embed.add_field(
        name="URL",
        value=f"{get_site_url()}/fittings/fit/{fit_id}/",
        inline=False
    )
    return embed


def doctrine_details(doctrine_id):
    try:
        doctrine = Doctrine.objects.get(id=doctrine_id)
    except ObjectDoesNotExist:
        embed = Embed(title="Doctrine Not Found")
        return embed

    embed = Embed(title=doctrine.name)
    embed.set_thumbnail(
        url=doctrine.icon_url
    )
    embed.description = doctrine.description

    for fit in doctrine.fittings.all():
        embed.add_field(
            name=fit.ship_type.name,
            value=f"[{fit.name}]({get_site_url()}/fittings/fit/{fit.id}/)"
        )

    embed.add_field(
        name="URL",
        value=f"{get_site_url()}/fittings/doctrine/{doctrine_id}/",
        inline=False
    )

    return embed


def category_details(category_id):
    try:
        category = Category.objects.get(id=category_id)
    except ObjectDoesNotExist:
        embed = Embed(title="Category Not Found")
        return embed

    embed = Embed(title=category.name)

    for doctrine in category.doctrines.all():
        embed.add_field(
            name=doctrine.name,
            value=f"[{doctrine.name}]({get_site_url()}/fittings/doctrine/{doctrine.id}/)"
        )
    embed.add_field(
        name="URL",
        value=f"{get_site_url()}/fittings/cat/{category_id}/",
        inline=False
    )

    return embed


def setup(bot):
    bot.add_cog(Fittings(bot))



from django.forms import ModelForm
from django.forms.widgets import TextInput

from .models import Category


# Create your forms here.
class CategoryForm(ModelForm):
    class Meta:
        model = Category
        fields = '__all__'
        widgets = {
            'color': TextInput(attrs={'type': 'color'}),
        }



from django.core.management import call_command
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Preloads common ship and module types from ESI."

    def handle(self, *args, **kwargs):
        call_command(
            "eveuniverse_load_types",
            "Fittings",
            "--category_id_with_dogma",
            "8",
            "--category_id_with_dogma",
            "7",
            "--category_id_with_dogma",
            "6",
            "--category_id_with_dogma",
            "18",
            "--category_id_with_dogma",
            "20",
            "--category_id_with_dogma",
            "22",
            "--category_id_with_dogma",
            "87"
        )



# Generated by Django 2.2.14 on 2020-08-17 20:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0012_typehistory'),
    ]

    operations = [
        migrations.AlterField(
            model_name='doctrine',
            name='description',
            field=models.TextField(max_length=1000),
        ),
    ]



# Generated by Django 4.0.7 on 2022-09-29 21:15

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0007_evetype_description'),
        ('fittings', '0015_fitting_created_fitting_last_updated_and_more'),
    ]

    operations = [
        migrations.DeleteModel(
            name='DogmaAttribute',
        ),
        migrations.DeleteModel(
            name='DogmaEffect',
        ),
        migrations.RemoveField(
            model_name='itemgroup',
            name='category',
        ),
        migrations.DeleteModel(
            name='ServerVersion',
        ),
        migrations.RemoveField(
            model_name='type',
            name='group',
        ),
        migrations.DeleteModel(
            name='TypeHistory',
        ),
        migrations.AlterField(
            model_name='fitting',
            name='ship_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='eveuniverse.evetype'),
        ),
        migrations.AlterField(
            model_name='fittingitem',
            name='type_fk',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='eveuniverse.evetype'),
        ),
        migrations.DeleteModel(
            name='ItemCategory',
        ),
        migrations.DeleteModel(
            name='ItemGroup',
        ),
        migrations.DeleteModel(
            name='Type',
        ),
    ]



from django.template.defaulttags import register


@register.filter()
def get_item(dict_, key):
    return dict_.get(key)


@register.filter(name='empty_slots')
def empty_slots(fittings: dict, slot_type: str):
    keys = fittings.keys()

    for key in keys:
        if slot_type in key:
            # This slot type is not empty
            return False
    # This slot type is completely empty
    return True


@register.filter
def break_html_lines(value):
    return value.replace("<br>", "\n").replace("<br />", "\n")



from concurrent.futures import ThreadPoolExecutor, as_completed

from allianceauth.services.hooks import get_extension_logger
from celery import shared_task
from eveuniverse.models import EveType, EveEntity

from .models import Fitting, FittingItem
from .providers import esi

logger = get_extension_logger(__name__)


class EftParser:
    def __init__(self, eft_text):
        self.eft_lines = eft_text.strip().splitlines()
    
    def parse(self):  
        # Remove /OFFLINE mentions due to pyfa when an offlined module is exported
        # Add fitting notes that the module is offlined
        parsed_fitting_notes = _removeOfflinedModulesMention(self.eft_lines)
        sections = []
        for section in _importSectionIter(parsed_fitting_notes['eft_lines']):
            sections.append(section)

        modules = []
        cargo = []
        drone_bay = []
        fighter_bay = []
        ship_type = ''
        fit_name = ''
        counter = 0  # Slot flag number
        last_line = ''
        
        for section in sections:
            counter = 0
            if section.isDroneBay():
                for line in section.lines:
                    quantity = line.split()[-1]
                    item_name = line.split(quantity)[0].strip()
                    drone_bay.append({'name': item_name, 'quantity': int(quantity.strip('x')), 'section_name': 'DroneBay'})
            elif section.isFighterBay():
                for line in section.lines:
                    quantity = line.split()[-1]
                    item_name = line.split(quantity)[0].strip()
                    fighter_bay.append({'name': item_name, 'quantity': int(quantity.strip('x')), 'section_name': 'FighterBay'})
            else:
                for line in section.lines:
                    if line.startswith('['):
                        if ',' in line:
                            ship_type, fit_name = line[1:-1].split(',')
                            continue

                        if 'empty' in line.strip('[]').lower():
                            continue
                    else:
                        if ',' in line:
                            module, charge = line.split(',')
                            modules.append({'name': module, 'charge': charge.strip(), 'count': counter})
                        else:
                            
                            quantity = line.split()[-1]  # Quantity will always be the last element, if it is there.
                            
                            if 'x' in quantity and quantity[1:].isdigit():
                                item_name = line.split(quantity)[0].strip()
                                cargo.append({'name': item_name, 'quantity': int(quantity.strip('x')), 'section_name': 'Cargo'})
                            else:
                                modules.append({'name': line.strip(), 'charge': '', 'count': counter})
                    counter += 1

        return {'ship': ship_type, 'name': fit_name, 'modules': modules, 'cargo': cargo, 'drone_bay': drone_bay,
                'fighter_bay': fighter_bay, 'fitting_notes': parsed_fitting_notes['fitting_notes']}



def _importSectionIter(lines):
    section = Section()
    for line in lines:
        if not line:
            if section.lines:
                yield section
                section = Section()
        else:
            section.lines.append(line)
    if section.lines:
        yield section


def _removeOfflinedModulesMention(lines):
    fitting_notes = ''
    eft_lines = []
    for line in lines:
        if '/OFFLINE' in line:
            line = line.replace(' /OFFLINE', '')
            if ',' in line:
                item_name = line.split(',')[0].strip()
                fitting_notes += '{} is offlined \n'.format(item_name)
            else:
                quantity = line.split()[-1]
                if 'x' in quantity and quantity[1:].isdigit():
                    item_name = line.split(quantity)[0].strip()
                    fitting_notes += '{} is offlined \n'.format(item_name)
                else:
                    item_name = line.strip()
                    fitting_notes += '{} is offlined \n'.format(item_name)
        eft_lines.append(line)
    return {'fitting_notes': fitting_notes, 'eft_lines': eft_lines}


class Section:
    def __init__(self):
        self.lines = []

    def isDroneBay(self):
        types = []
        for line in self.lines:
            if line.startswith('['):
                return False
            if ',' in line:
                types.append(_get_type(line.split(',')[0].strip()))
            else:
                quantity = line.split()[-1]
                if 'x' in quantity and quantity[1:].isdigit():
                    types.append(_get_type(line.split(quantity)[0].strip()))
                else:
                    types.append(_get_type(line.strip()))
        return all(_type is not None and _type.eve_group.eve_category.id == 18 for _type in types)

    def isFighterBay(self):
        types = []
        for line in self.lines:
            if line.startswith('['):
                return False
            if ',' in line:
                types.append(_get_type(line.split(',')[0].strip()))
            else:
                quantity = line.split()[-1]
                if 'x' in quantity and quantity[1:].isdigit():
                    types.append(_get_type(line.split(quantity)[0].strip()))
                else:
                    types.append(_get_type(line.strip()))
        return all(_type is not None and _type.eve_group.eve_category.id == 87 for _type in types)


def _get_type(type_name):
    try:
        type_obj = EveType.objects.get(name=type_name)

        return type_obj
    except:
        # If the type is not already in the db, then we have to resolve type_id before we have eveuniverse
        # create the object.
        c = esi.client
        try:
            type_id = c.Universe.post_universe_ids(names=[type_name]).result()['inventory_types'][0]["id"]
            type_obj = EveType.objects.get_or_create_esi(id=type_id, enabled_sections=[EveType.Section.DOGMAS])
            type_obj = type_obj[0]

            return type_obj
        except TypeError as e:
            logger.error(f"Type ID for {type_name} not returned via ESI. Consider removing this from your fit.")
            raise e


@shared_task()
def create_fitting_item(fit, item):
    count = None
    quantity = None
    if 'count' in item:
        count = item['count']

    type_obj = _get_type(item['name'])

    # Dogma Effects
    flags = {11: 'LoSlot', 12: 'HiSlot', 13: 'MedSlot', 2663: 'RigSlot', 3772: 'SubSystemSlot', 6306: 'ServiceSlot'}
    effects = type_obj.dogma_effects.filter(eve_dogma_effect_id__in=flags).values_list('eve_dogma_effect_id', flat=True)
    effects = list(effects)
    if count is None:
        flag = item['section_name']
        quantity = item['quantity']
    # Check due to active drug from pyfa not showing quantities
    elif len(effects) == 0:
        flag = 'Cargo'
        quantity = 1
    else:
        flag = flags[effects[0]] + str(count)

    item = FittingItem.objects.create(flag=flag, quantity=quantity if quantity else 1, type_fk=type_obj,
                                      type_id=type_obj.pk, fit=fit)


@shared_task
def create_fit(eft_text, description=None):
    parsed_eft = EftParser(eft_text).parse()
    if description is None or len(description) == 0:
        description += '{}'.format(parsed_eft['fitting_notes'])
    else:
        description += '\n {}'.format(parsed_eft['fitting_notes'])

    logger.info("Creating fit.")
    logger.debug(f"Fit name: {parsed_eft['name']}, Type: {parsed_eft['ship']}")

    def __create_fit(ship_type, name, description):
        type_obj = _get_type(ship_type)
        if name == " " or name == "":
            name = "Unnamed " + ship_type + " fitting"
        fit = Fitting.objects.create(ship_type=type_obj, ship_type_type_id=type_obj.pk,
                                     name=name, description=description)
        return fit

    fit = __create_fit(parsed_eft['ship'], parsed_eft['name'], description)
    create_fitting_items(fit, parsed_eft)

    logger.info("Done creating fit.")


@shared_task 
def create_fitting_items(fit, parsed_eft):
    type_names = [x['name'] for x in parsed_eft['modules']]
    type_names += [x['name'] for x in parsed_eft['cargo']]
    type_names += [x['name'] for x in parsed_eft['drone_bay']]
    type_names += [x['name'] for x in parsed_eft['fighter_bay']]
    type_names = list(set(type_names))

    # Get a list of types missing from the db
    types = EveType.objects.filter(name__in=type_names).values_list('name', flat=True)

    missing = [x for x in type_names if x not in types]

    # Create missing types
    ids = EveEntity.objects.fetch_by_names_esi(missing)\
        .filter(category=EveEntity.CATEGORY_INVENTORY_TYPE)\
        .values_list("id", flat=True)
    _ = EveType.objects.bulk_get_or_create_esi(ids=ids)

    # Create the fitting items
    for module in parsed_eft['modules']:
        create_fitting_item(fit, module)

    for item in parsed_eft['cargo']:
        create_fitting_item(fit, item)

    for item in parsed_eft['drone_bay']:
        create_fitting_item(fit, item)

    for item in parsed_eft['fighter_bay']:
        create_fitting_item(fit, item)


@shared_task
def update_fit(eft_text, fit_id, description=None):
    parsed_eft = EftParser(eft_text).parse()
    fit = Fitting.objects.get(id=fit_id)

    if parsed_eft['ship'] != fit.ship_type.name:
        logger.info("Cannot update a fitting with different ship type")
        return

    logger.info("Updating Fit name: {parsed_eft['name']}, Type: {parsed_eft['ship']}")  

    FittingItem.objects.filter(fit__id=fit_id).delete() 

    create_fitting_items(fit, parsed_eft)

    fit.name = parsed_eft['name']
    fit.description = description
    fit.save()

    logger.info("Done updating fit " + fit_id)




from django.urls import re_path

from . import views

app_name = 'fittings'

urlpatterns = [
    re_path(r'^$', views.dashboard, name='dashboard'),
    re_path(r'^fit/add/$', views.add_fit, name='add_fit'),
    re_path(r'^fit/all/$', views.view_all_fits, name='view_all_fits'),
    re_path(r'^fit/(?P<fit_id>[0-9]+)/$', views.view_fit, name='view_fit'),
    re_path(r'^fit/(?P<fit_id>[0-9]+)/delete/$', views.delete_fit, name='delete_fit'),
    re_path(r'^fit/(?P<fit_id>[0-9]+)/save/$', views.save_fit, name='save_fit'),
    re_path(r'^fit/(?P<fit_id>[0-9]+)/edit/$', views.edit_fit, name='edit_fit'),
    re_path(r'^doctrine/add/$', views.add_doctrine, name='add_doctrine'),
    re_path(r'^doctrine/(?P<doctrine_id>[0-9]+)/$', views.view_doctrine, name='view_doctrine'),
    re_path(r'^doctrine/(?P<doctrine_id>[0-9]+)/delete/$', views.delete_doctrine, name='delete_doctrine'),
    re_path(r'^doctrine/(?P<doctrine_id>[0-9]+)/edit/$', views.edit_doctrine, name='edit_doctrine'),
    re_path(r'^cat/all/$', views.view_all_categories, name='view_all_categories'),
    re_path(r'^cat/add/$', views.add_category, name='add_category'),
    re_path(r'^cat/(?P<cat_id>[0-9]+)/$', views.view_category, name='view_category'),
    re_path(r'^cat/(?P<cat_id>[0-9]+)/edit/$', views.edit_category, name='edit_category'),
    re_path(r'^cat/(?P<cat_id>[0-9]+)/delete/$', views.delete_category, name='delete_category'),
]
